import torch
from torch.autograd        import grad
from sparse_distance_bmp   import sparse_distance_bmp, extract_point_cloud
from pykeops.torch         import Kernel

import shapes_toolbox as st
from shapes_toolbox import PointCloud
from shapes_toolbox.models import fit_model, SmallDeformation, LddmmGeodesic


import numpy as np
from matplotlib import pyplot as plt
from scipy import misc
from time  import time

use_cuda = torch.cuda.is_available()
tensor   = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor 

plt.ion()
plt.show()

# Choose the storage place for our data : CPU (host) or GPU (device) memory.
use_cuda = torch.cuda.is_available()
dtype    = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor
dtypeint = torch.cuda.LongTensor  if use_cuda else torch.LongTensor

# Make sure that everybody's on the same wavelength:
st.dtype = dtype ; st.dtypeint = dtypeint

s2v = lambda x : tensor([x]).type(dtype)


from scipy.ndimage.filters import gaussian_filter

def LoadImage(fname) :
    img = misc.imread(fname, flatten = True) # Grayscale
    #img = gaussian_filter(img, 1, mode='nearest')
    img = (img[::-1, :])  / 255.
    img = np.swapaxes(img, 0,1 )
    return tensor( 1 - img )

source = LoadImage("data/cerberus_a.png")
target = LoadImage("data/cerberus_b.png")

source = LoadImage("data/circle_a.png")
target = LoadImage("data/circle_b.png")
# source = LoadImage("data/worm_c.png")
# target = LoadImage("data/worm_d.png")

def subsample(x) :
    return .25 * (x[::2, ::2] + x[1::2, ::2] + x[::2, 1::2] + x[1::2, 1::2])
    
source = subsample(subsample(source))
target = subsample(subsample(target))

# The images are rescaled to fit into the unit square ==========================================
scale = source.shape[0]
affine = tensor( [[ 1, 0, 0 ],
                  [ 0, 1, 0 ] ]) / scale

_, a_i, x_i = extract_point_cloud(source, affine)
_, b_j, y_j = extract_point_cloud(target, affine)

a_i = a_i / a_i.sum()
b_j = b_j / b_j.sum()

alpha = PointCloud(x_i, a_i)
beta  = PointCloud(y_j, b_j)

params = {
	"weight_deformation" : .1,
	"weight_fidelity"    : 10.,

	"deformation" : {
		"id"         : Kernel("cauchy(x,y)"),
		"gamma"      :             s2v(1/.1**2),
	},

	"display" : {
		"limits"     : (0,1,0,1),
		"info"       : True,
		"grid"       : True,
        "grid_color" :    (.9,.9,.9),

        "template"   : False,
        "model_color":    (.8, 0., 0.),

        "template_linewidth" : 10,
        "model_linewidth" : 10,
        "target_linewidth" : 10,
        "model_gradient_scale" : .2,
        "model_gradient_color":  (0., .7, 0.),
        "info"       : True,
        "movie"      : True,
	},
    "optimization" : {
        "nits"   : 100,
        "nlogs"  : 10,
    },
	"save" : {
		"output_directory"   : "output/registration/",
	},
}

if False :
	params["fidelity"] = {
		"formula"    : "kernel",
		"features"   : "locations",
		"id"         : Kernel("-distance(x,y)"),
		"gamma"      :       s2v(1.),
	}

elif False : 
    params["fidelity"] = {
        "formula"    : "sinkhorn",
        "epsilon"     : s2v( .01 ),
        "kernel"      : {
            "id"     : Kernel("gaussian(x,y)") ,
            "gamma"  :  1 / s2v( .01 )      },
        "tol"         : 1e-6,
        "nits"        : 30,
        "rho"         : -1., # Balanced transport. Make sure that both measures have equal mass!
        "transport_plan" : "minimal_symmetric+heatmaps",
    }

else : 
    params["fidelity"] = {
        "formula"    : "wasserstein",
        "epsilon"     : s2v( .2**2 ),
        "kernel"      : {
            "id"     : Kernel("gaussian(x,y)") ,
            "gamma"  :  1 / s2v( .2**2 )      },
        "tol"         : 1e-6,
        "nits"        : 30,
        "rho"         : -1., # Balanced transport. Make sure that both measures have equal mass!
        "transport_plan" : "minimal_symmetric",
    }



# Define our (simplistic) matching model
deformation = "TPS"
if   deformation == "LDDMM" : # LDDMM
	model = LddmmGeodesic(alpha)
elif deformation == "kernel" :    # Simple tangential deformation
	model = SmallDeformation(alpha)
elif deformation == "TPS" :    # Thin plate spline
    #params["deformation"] = {
    #    "id"         : Kernel("thin_plate_spline(x,y)"),
    #    "gamma"      :                s2v(1.),
    #}
    model = SmallDeformation(alpha, affine_strength=10)


# Train it
fit_model(params, model, beta)
